import ast
import functools
import inspect
from inspect import getmembers, getmodule, getsource, isbuiltin
from textwrap import dedent
from types import FunctionType, MethodDescriptorType, MethodType, MethodWrapperType
from typing import TYPE_CHECKING, Any, List, Optional, Union

if TYPE_CHECKING:
    _Base = ast.NodeVisitor
else:
    _Base = object


class DeepMixin(_Base):
    def __init__(self) -> None:
        self.module = None
        self.obj = None
        self.visited_nodes = 0
        self.raw_nodes: List[str] = []
        self.parent_nodes: List[str] = []
        self.last_obj: Union[None, FunctionType, MethodType, object] = None
        super().__init__()

    def deep_visit(self, callable: Union[FunctionType, MethodType]):
        """Visit all AST nodes of the passed in `callable`. This will include the AST of any [ast.Call](https://docs.python.org/3/library/ast.html#ast.Call) nodes that are encountered.

        Args:
            callable (Union[FunctionType, MethodType]): The function or method that will be "deep" visited.
        """

        start_node = None

        self.module = getmodule(callable)  # type: ignore

        parent = get_class_that_defined_method(callable)

        if parent:
            self.obj = parent
            self.last_obj = parent

        start_node = self._convert_to_ast_node(callable)

        if not start_node:
            raise Exception(f"Could not find AST node for {callable}")

        self.visit(start_node)

    def _convert_to_ast_node(
        self,
        item: Union[FunctionType, MethodType, MethodWrapperType],
        record_node: bool = True,
    ) -> Optional[ast.AST]:

        if isinstance(item, MethodWrapperType):
            # print(f"Unable to process method wrapper: {item.__name__}")
            return None

        if isinstance(item, MethodDescriptorType):
            # print(f"Unable to process method desciption: {item.__name__}")
            return None

        if record_node:
            self._record_node(item)

        try:
            source = self._get_source(item)
        except TypeError:
            # print(f"Invalid type {type(item)} for {item.__name__}")
            return None

        return ast.parse(source)

    def _record_node(self, item: Union[FunctionType, MethodType]):

        if isinstance(item, MethodType):
            self.last_obj = item.__self__
            self.parent_nodes.append(
                f"{item.__self__.__class__.__name__}.{item.__name__}()"
            )
            return

        if isinstance(item, FunctionType):

            parent_class = get_class_that_defined_method(item)

            if parent_class:
                self.last_obj = parent_class
                self.parent_nodes.append(f"{parent_class.__name__}.{item.__name__}()")
                return

            self.parent_nodes.append(f"{item.__name__}()")
            return

        if isinstance(item, type):
            self.parent_nodes.append(f"{item.__name__}.init()")
            return

    def _get_source(self, item: Any):
        return dedent(getsource(item))

    def _module_search(self, item: str, excludes: Optional[List] = None):

        # print(f"Searching for {item}")

        if excludes is None:
            excludes = []

        for name, object in getmembers(self.module):

            if object in excludes:
                continue

            if name == item:

                # For some reason built-in exceptions
                # get pass this check.
                if isbuiltin(object):
                    print(f"Skipping built-in {name}")
                    return None

                return object

            attr = getattr(object, item, None)

            if attr is not None:

                if isbuiltin(attr):
                    print(f"Skipping built-in {name}")
                    return None

                return attr

    def _find_ast_node(self, item: str):

        item_object = self._module_search(item, [self.last_obj])

        if item_object is None:
            return item_object

        return self._convert_to_ast_node(item_object)

    def visit(self, node: ast.AST) -> Any:

        self.visited_nodes += 1

        self.raw_nodes.append(node.__class__.__name__)

        if isinstance(node, ast.Call):
            self._proccess_call(node)

        # if isinstance(node, ast.ClassDef):
        #     self._process_class_def(node)

        return super().visit(node)

    def _process_attr(self, node: ast.Attribute):

        obj_name: Optional[str] = None
        method_name: str = node.attr

        # like self.name
        if isinstance(node.value, ast.Attribute):
            return

        # like self.speak()
        if isinstance(node.value, ast.Name):
            obj_name = node.value.id

        # like super().speak()
        if isinstance(node.value, ast.Call):

            # todo fix this
            # like foo().bar().bazz()
            if isinstance(node.value.func, ast.Attribute):
                # print("Found nested attribute")
                # print(ast.dump(node, indent=4))
                return

            method_name = node.attr
            obj_name = node.value.func.id  # type: ignore[attr-defined]

            if obj_name == "super":
                self._process_super(method_name)
                return

        # Self doesnt always mean the object we started with... We could be parsing a method on some other
        # class than which we started.
        if obj_name == "self":
            method_obj = getattr(self.last_obj, method_name, None)

            if method_obj is not None:

                method_node = self._convert_to_ast_node(method_obj)

                if not method_node:
                    print(f"Unable to find method object for {obj_name}.{method_obj}()")
                    return

                self.visit(method_node)
                return

            # Everthing below this line should not be in the "if" statement...
            # but doing so makes the problem worse. The problem is with var.foo()
            # how do you know what class "var" is so you can look up the method "foo".
            # Right now I dont know but its something I'm going to work on...

            # If method_object is None we should still search the module tree for
            # everything but the current obj and start object, this is still a bad approach
            # because there could still be duplicate objects with the same method todo list to fix

            method_node = self._find_ast_node(method_name)

            if not method_node:
                # print(f"YOLO search unable to find {method_name} in {self.module}")
                return

            self.visit(method_node)
            return

    def _get_class_def(self, item: object):
        class_node = self._convert_to_ast_node(item, record_node=False)  # type: ignore

        if class_node is None:
            raise Exception(f"Unable to get AST node for {item}")

        class_def = _find_class_def(class_node)

        if class_def is None:
            Exception(f"Failed to find ClassDef for {item}")

        return class_def

    def _process_super(self, method_name: str):

        class_def = self._get_class_def(self.last_obj)

        parent_class_names = []

        for parent in class_def.bases:
            if isinstance(parent, ast.Attribute):
                parent_class_names.append(parent.attr)
                continue

            parent_class_names.append(parent.id)

        if not parent_class_names:
            print(f"Failed to find parent classes for {method_name}")
            return

        # print(parent_class_names)

        parent_search_results = [
            self._module_search(class_name) for class_name in parent_class_names
        ]

        parent_class_objs = [
            parent_obj for parent_obj in parent_search_results if parent_obj is not None
        ]

        if not parent_class_objs:
            print(f"Failed to convert {parent_class_names} to {parent_class_objs}")

        attr_obj = None

        for parent_obj in parent_class_objs:
            attr_obj = getattr(parent_obj, method_name, None)

            if attr_obj:
                break

        if attr_obj is None:
            print(f"Failed to find {method_name} in {parent_class_objs}")
            print(f"{parent_class_objs}")
            return

        attr_node = self._convert_to_ast_node(attr_obj)

        if attr_node is None:
            return

        self.visit(attr_node)

        return

    def _proccess_name(self, node: ast.Name):
        func_name = node.id

        # There is bug for super().foo()
        # For what ever reason it seems like the attribute foo()
        # comes first in the AST and then super().
        if func_name == "super":
            return

        func_node = self._find_ast_node(func_name)

        if not func_node:
            # print(f"Unable to find {func_name} in {self.module.__name__}")
            return

        self.visit(func_node)

    def _proccess_call(self, node: ast.Call):

        if isinstance(node.func, ast.Attribute):
            self._process_attr(node.func)
            return

        if isinstance(node.func, ast.Name):
            self._proccess_name(node.func)
            return


def _find_class_def(start_node: ast.AST):
    for node in ast.walk(start_node):
        if isinstance(node, ast.ClassDef):
            return node


# Thanks Stackoverflow! https://stackoverflow.com/a/25959545/3888850
def get_class_that_defined_method(meth):
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (
        inspect.isbuiltin(meth)
        and getattr(meth, "__self__", None) is not None
        and getattr(meth.__self__, "__class__", None)
    ):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, "__func__", meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(
            inspect.getmodule(meth),
            meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0],
            None,
        )
        if isinstance(cls, type):
            return cls
    return getattr(meth, "__objclass__", None)  # handle special descriptor objects



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['deep_ast']

package_data = \
{'': ['*']}

install_requires = \
['click-log>=0.4.0,<0.5.0', 'click>=8.1.2,<9.0.0']

entry_points = \
{'console_scripts': ['deep-ast = deep_ast.app:cli']}

setup_kwargs = {
    'name': 'deep-ast',
    'version': '0.0.1',
    'description': 'Walks the entire AST tree, including calls to other functions.',
    'long_description': '<!-- PROJECT SHIELDS -->\n<!--\n*** I\'m using markdown "reference style" links for readability.\n*** Reference links are enclosed in brackets [ ] instead of parentheses ( ).\n*** See the bottom of this document for the declaration of the reference variables\n*** for contributors-url, forks-url, etc. This is an optional, concise syntax you may use.\n*** https://www.markdownguide.org/basic-syntax/#reference-style-links\n-->\n[![Python][python-shield]][pypi-url]\n[![Latest][version-shield]][pypi-url]\n[![Tests][test-shield]][test-url]\n[![Coverage][codecov-shield]][codecov-url]\n[![License][license-shield]][license-url]\n<!-- [![Contributors][contributors-shield]][contributors-url]\n[![Forks][forks-shield]][forks-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url] -->\n\n<!-- PROJECT LOGO -->\n<br />\n<p align="center">\n  <!-- <a href="https://github.com/DontShaveTheYak/deep-ast">\n    <img src="images/logo.png" alt="Logo" width="80" height="80">\n  </a> -->\n\n  <h3 align="center">Deep-AST</h3>\n\n  <p align="center">\n    Parse the AST for any <a href=https://docs.python.org/3/library/ast.html#ast.Call>Call</a> nodes encountered.\n    <!-- <br />\n    <a href="https://github.com/DontShaveTheYak/deep-ast"><strong>Explore the docs »</strong></a>\n    <br /> -->\n    <br />\n    <!-- <a href="https://github.com/DontShaveTheYak/deep-ast">View Demo</a>\n    · -->\n    <a href="https://github.com/DontShaveTheYak/deep-ast/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/DontShaveTheYak/deep-ast/issues">Request Feature</a>\n    ·\n    <!-- <a href="https://la-tech.co/post/hypermodern-cloudformation/getting-started/">Guide</a> -->\n  </p>\n</p>\n\n## About The Project\n\n<!-- [![Product Name Screen Shot][product-screenshot]](https://example.com) -->\n\n`deep-ast` is a Python library that attempts to parse all AST that would be encountered when invoking a callable in python. It does this by extending the [ast.NodeVistor](https://docs.python.org/3/library/ast.html#ast.NodeVisitor) and [ast.NodeTransformer](https://docs.python.org/3/library/ast.html#ast.NodeTransformer) classes, so that when an [ast.Call](https://docs.python.org/3/library/ast.html#ast.Call) node is encountered, it\'s source code is parsed into an [ast.Node](https://docs.python.org/3/library/ast.html#ast.Call) and then passes the node to the [visit()](https://docs.python.org/3/library/ast.html#ast.NodeVisitor.visit) method.\n\nWe say attempts because currently there are some limitations. It cant parse any arbitrary code like `exec(\'rorrEeulaV esiar\'[::-1])` which raises a ValueError, functions that aren\'t written in Python and all python internals like `print()`.\n\nIf you have a way around these limitations then a PR would be greatly appreciated.\n\n## Getting Started\n\n### Prerequisites\n\n`deep-ast` requires python >= 3.7\n\n### Installation\n\n`deep-ast` is available as an easy to install pip package.\n```sh\npip install deep-ast\n```\n\n## Usage\n\ndeep-ast offers drop in replacements for the [ast.NodeVistor](https://docs.python.org/3/library/ast.html#ast.NodeVisitor) and [ast.NodeTransformer](https://docs.python.org/3/library/ast.html#ast.NodeTransformer) classes.\n\n```sh\nfrom deep_ast import DeepVisitor, DeepTransformer\n```\n\nTo start the deep processing you the `deep_visit()` method. Each function takes an optional `module` arugment. This is used if the function/method is not defined in the same file that you calling `deep_visit()` method in.\n\n### Example to get all exceptions\n\nThis example shows how you might deeply parse the AST to\nget all exceptions that might be thrown.\n\n```python3\n# Custom NodeVisitor to visit Raise nodes and record them\nclass ParseExceptions(DeepVisitor):\n    def __init__(self) -> None:\n        self.raw_exceptions = []\n        self.found_exceptions = []\n        super().__init__()\n\n    def _add_exception(self, name: str):\n        self.raw_exceptions.append(name)\n\n        if name not in self.found_exceptions:\n            self.found_exceptions.append(name)\n\n    def visit_Raise(self, node):\n\n        exception_obj = node.exc\n\n        if isinstance(exception_obj, (ast.Call, ast.Name):\n            name = (\n                exception_obj.id\n                if isinstance(exception_obj, ast.Name)\n                else exception_obj.func.id\n            )\n\n            self._add_exception(name)\n            return self.generic_visit(node)\n\n        self._add_exception("EmptyRaise")\n        return self.generic_visit(node)\n\n# Test functions to visit\ndef foo():\n    bar()\n    raise TypeError()\n\n\ndef bar():\n    raise ValueError()\n\nparser = ParseExceptions()\n\nparser.deep_visit(foo)\n\nprint(parser.found_exceptions) # prints [\'ValueError\', \'TypeError\']\n```\n\n## Roadmap\n\n- Parsing of deeply nested attribute calls like `foo().bar().bazz()`\n\n- Tracking of variable assignment:\n  ```python3\n  print_func = print\n  print_func()\n  ```\n\nSee the [open issues](https://github.com/DontShaveTheYak/deep-ast/issues) for a list of proposed features (and known issues).\n\n## Contributing\n\nContributions are what make the open-source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nThis project uses poetry to manage dependencies and pre-commit to run formatting, linting and tests. You will need to have both installed to your system as well as python 3.9.\n\n1. Fork the Project\n2. Setup the environment.  \n   This project uses vscode devcontainer to provide a completly configured development environment. If you are using vscode and have the remote container extension installed, you should be asked to use the devcontainer when you open this project inside of vscode.\n\n   If you are not using devcontainers then you will need to have python installed. Install the `poetry`, `nox`, `nox_poetry` and `pre-commit` packages. Then run `poetry install` and `pre-commit install` commands. \n\n   Most of the steps can be found in the [Dockerfile](.devcontainer/Dockerfile).\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n<!-- ## License\n\nDistributed under the Apache-2.0 License. See [LICENSE.txt](./LICENSE.txt) for more information. -->\n\n## Contact\n\nLevi - [@shady_cuz](https://twitter.com/shady_cuz)\n\n<!-- ACKNOWLEDGEMENTS -->\n## Acknowledgements\n* Stackoverflow for getting me started [down this road](https://stackoverflow.com/questions/32560116/how-to-list-all-exceptions-a-function-could-raise-in-python-3).\n* @sethmlarson for asking me which [exceptions](https://github.com/urllib3/urllib3/issues/2648) can be raised.\n\n<!-- MARKDOWN LINKS & IMAGES -->\n<!-- https://www.markdownguide.org/basic-syntax/#reference-style-links -->\n[python-shield]: https://img.shields.io/pypi/pyversions/deep-ast?style=for-the-badge\n[version-shield]: https://img.shields.io/pypi/v/deep-ast?label=latest&style=for-the-badge\n[pypi-url]: https://pypi.org/project/deep-ast/\n[test-shield]: https://img.shields.io/github/workflow/status/DontShaveTheYak/deep-ast/Tests?label=Tests&style=for-the-badge\n[test-url]: https://github.com/DontShaveTheYak/deep-ast/actions?query=workflow%3ATests+branch%3Amaster\n[codecov-shield]: https://img.shields.io/codecov/c/gh/DontShaveTheYak/deep-ast/master?color=green&style=for-the-badge&token=bfF18q99Fl\n[codecov-url]: https://codecov.io/gh/DontShaveTheYak/deep-ast\n[contributors-shield]: https://img.shields.io/github/contributors/DontShaveTheYak/deep-ast.svg?style=for-the-badge\n[contributors-url]: https://github.com/DontShaveTheYak/deep-ast/graphs/contributors\n[forks-shield]: https://img.shields.io/github/forks/DontShaveTheYak/deep-ast.svg?style=for-the-badge\n[forks-url]: https://github.com/DontShaveTheYak/deep-ast/network/members\n[stars-shield]: https://img.shields.io/github/stars/DontShaveTheYak/deep-ast.svg?style=for-the-badge\n[stars-url]: https://github.com/DontShaveTheYak/deep-ast/stargazers\n[issues-shield]: https://img.shields.io/github/issues/DontShaveTheYak/deep-ast.svg?style=for-the-badge\n[issues-url]: https://github.com/DontShaveTheYak/deep-ast/issues\n[license-shield]: https://img.shields.io/github/license/DontShaveTheYak/deep-ast.svg?style=for-the-badge\n[license-url]: https://github.com/DontShaveTheYak/deep-ast/blob/master/LICENSE.txt\n[product-screenshot]: images/screenshot.png\n',
    'author': 'Levi Blaney',
    'author_email': 'shadycuz@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/DontShaveTheYak/deep-ast',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
